﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using HeuristicLab.Services.Optimization.ControllerService.Interfaces;
using Microsoft.WindowsAzure;
using Microsoft.WindowsAzure.StorageClient;

namespace HeuristicLab.Services.Optimization.ControllerService.Azure {
  public static class AzureConstants {
    public static readonly string SCENARIO_TABLE = "Scenario";
    public static readonly string SCENARIO_BLOB_CONTAINER = "scenario";
    public static readonly string CLOUD_SETTINGS_KEY = "Cloudia.WindowsAzure.Storage";
  }

  public class ScenarioDao : IScenarioDao {    
    public CloudTableClient TableClient { get; set; }

    public bool Add(ScenarioEntity entity) {
      TableServiceContext serviceContext = TableClient.GetDataServiceContext();
      TableClient.CreateTableIfNotExist(AzureConstants.SCENARIO_TABLE);
      ScenarioEntity dbEntity = (from e in serviceContext.CreateQuery<ScenarioEntity>(AzureConstants.SCENARIO_TABLE)
                               where e.RowKey == entity.RowKey
                               select e).FirstOrDefault();
      if (dbEntity != null)
        return false;

      serviceContext.AddObject(AzureConstants.SCENARIO_TABLE, entity);
      serviceContext.SaveChanges();
      return true;
    }

    public bool DeleteByName(string scenarioName) {
      TableServiceContext serviceContext = TableClient.GetDataServiceContext();
      TableClient.CreateTableIfNotExist(AzureConstants.SCENARIO_TABLE);
      ScenarioEntity entity = (from e in serviceContext.CreateQuery<ScenarioEntity>(AzureConstants.SCENARIO_TABLE)
                               where e.RowKey == scenarioName
                               select e).FirstOrDefault();
      if (entity == null)
        return false;

      serviceContext.DeleteObject(entity);
      serviceContext.SaveChangesWithRetries();
      return true;
    }

    public ScenarioEntity FindByName(string scenarioName) {
      TableServiceContext serviceContext = TableClient.GetDataServiceContext();
      TableClient.CreateTableIfNotExist(AzureConstants.SCENARIO_TABLE);
      ScenarioEntity entity = (from e in serviceContext.CreateQuery<ScenarioEntity>(AzureConstants.SCENARIO_TABLE)
                               where e.RowKey == scenarioName
                               select e).FirstOrDefault();
      return entity;
    }


    public IEnumerable<ScenarioEntity> GetAllEntities() {
      TableServiceContext serviceContext = TableClient.GetDataServiceContext();
      TableClient.CreateTableIfNotExist(AzureConstants.SCENARIO_TABLE);
      return (from e in serviceContext.CreateQuery<ScenarioEntity>(AzureConstants.SCENARIO_TABLE)
             select e).AsEnumerable();
    }
  }

  public class BlobDao : IBlobDao {
    public CloudBlobClient BlobClient { get; set; }

    public bool Add(StringEntry entry) {
      CloudBlobContainer container = BlobClient.GetContainerReference(AzureConstants.SCENARIO_BLOB_CONTAINER);
      container.CreateIfNotExist();
      var blob = container.GetBlobReference(entry.Key);
      blob.UploadText(entry.Text);
      return true;
    }

    public bool DeleteByKey(string entryKey) {
      CloudBlobContainer container = BlobClient.GetContainerReference(AzureConstants.SCENARIO_BLOB_CONTAINER);
      container.CreateIfNotExist();
      var blob = container.GetBlobReference(entryKey);
      return blob.DeleteIfExists();
    }

    public StringEntry FindByKey(string entryKey) {
      CloudBlobContainer container = BlobClient.GetContainerReference(AzureConstants.SCENARIO_BLOB_CONTAINER);
      container.CreateIfNotExist();
      var blob = container.GetBlobReference(entryKey);
      return new StringEntry() { Key = entryKey, Text = blob.DownloadText() };      
    }
  }

  public class AzureDataAccessLayer : IDataAccessLayer {
    private IScenarioDao scenarioDao;
    private IBlobDao blobDao;

    private CloudStorageAccount storageAccount;

    private CloudStorageAccount StorageAccount {
      get {
        if (storageAccount == null) {
          storageAccount = CloudStorageAccount.Parse(CloudConfigurationManager.GetSetting(AzureConstants.CLOUD_SETTINGS_KEY));
        }
        return storageAccount; 
      }      
    }

    public IScenarioDao ScenarioDao {
      get {
        if (scenarioDao == null) {
          scenarioDao = new ScenarioDao();
        }
        return scenarioDao;
      }
    }

    public IBlobDao BlobDao {
      get {
        if (blobDao == null) {
          blobDao = new BlobDao();
        }
        return blobDao;
      }
    }

    public IScenarioDao CreateScenarioDao() {
      return new ScenarioDao() { TableClient = StorageAccount.CreateCloudTableClient() };
    }

    public IBlobDao CreateBlobDao() {
      return new BlobDao() { BlobClient = StorageAccount.CreateCloudBlobClient() };
    }
  }
}
